// Copyright 2000-2002 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OAInternetConfig.h"

#import <AppKit/NSPanel.h>
#import <OmniFoundation/OmniFoundation.h>
#import <Foundation/Foundation.h>
#import <CoreFoundation/CoreFoundation.h>
#import <Carbon/Carbon.h>
#import <OmniBase/OmniBase.h>

#import "OAOSAScript.h"

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/Frameworks/OmniAppKit/Carbon/OAInternetConfig.m,v 1.17 2002/03/09 01:53:53 kc Exp $")

// The InternetConfig documentation can be found at http://www.quinn.echidna.id.au/Quinn/Config/.

@interface OAInternetConfig (Private)
- (NSString *)_helperKeyNameForScheme:(NSString *)scheme;
static NSString *OANameForInternetConfigErrorCode(OSStatus errorCode);
@end

static NSString *helperKeyPrefix = nil;

@implementation OAInternetConfig

+ (void)initialize;
{
    OBINITIALIZE;

    helperKeyPrefix = (NSString *)CFStringCreateWithPascalString(NULL, kICHelper, kCFStringEncodingMacRoman);
}

+ (OAInternetConfig *)internetConfig;
{
    return [[[self alloc] init] autorelease];
}

+ (unsigned long)applicationSignature;
{
    NSString *bundleSignature;
    NSData *signatureBytes;

    bundleSignature = [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleSignature"];
    if (!bundleSignature)
        return FOUR_CHAR_CODE('????');
    
    signatureBytes = [bundleSignature dataUsingEncoding:[NSString defaultCStringEncoding]];
    if (!signatureBytes || ([signatureBytes length] < 4))
        return FOUR_CHAR_CODE('????');
    
    return *(const OSType *)[signatureBytes bytes];
}

// Init and dealloc

- init;
{
    ICInstance anInstance;

    if (![super init])
        return nil;
    if (ICStart(&anInstance, [[self class] applicationSignature]) != noErr)
        return nil;
    internetConfigInstance = anInstance;
    return self;
}

- (void)dealloc;
{
    ICStop(internetConfigInstance);
    [super dealloc];
}

//

- (OAInternetConfigInfo)getInfoForScheme:(NSString *)scheme;
{
    OAInternetConfigInfo info;
    NSString *helperKeyString;
    Str255 helperKeyPascalString;
    ICAppSpec helperApplication;
    ICAttr attribute;
    long helperApplicationSize;
    OSStatus error;

    helperKeyString = [self _helperKeyNameForScheme:scheme];
    CFStringGetPascalString((CFStringRef)helperKeyString, helperKeyPascalString, sizeof(helperKeyPascalString) - 1, kCFStringEncodingMacRoman);

    helperApplicationSize = sizeof(helperApplication);
    error = ICGetPref(internetConfigInstance, helperKeyPascalString, &attribute, &helperApplication, &helperApplicationSize);
    switch (error) {
        case noErr:
            info.valid = YES;
            info.applicationName = [(NSString *)CFStringCreateWithPascalString(NULL, helperApplication.name, kCFStringEncodingMacRoman) autorelease];
            info.creatorCode = helperApplication.fCreator;
            break;
        case icPrefNotFoundErr:
            info.valid = NO;
#ifdef DEBUG
            NSLog(@"OAInternetConfig: ICGetPref found no helper application for %@", scheme);
#endif
            break;
        default:
            info.valid = NO;
            NSLog(@"OAInternetConfig: ICGetPref returned an error while getting key %@: %@", helperKeyString, OANameForInternetConfigErrorCode(error));
            break;
    }
    
    return info;
}

- (NSString *)helperApplicationForScheme:(NSString *)scheme;
{
    OAInternetConfigInfo info;

    info = [self getInfoForScheme:scheme];
    return info.valid ? info.applicationName : nil;
}

- (void)setApplicationCreatorCode:(long)applicationCreatorCode name:(NSString *)applicationName forScheme:(NSString *)scheme;
{
    NSString *helperKeyString;
    Str255 helperKeyPascalString;
    ICAppSpec helperApplication;
    long helperApplicationSize;
    OSStatus error;

    helperKeyString = [self _helperKeyNameForScheme:scheme];
    CFStringGetPascalString((CFStringRef)helperKeyString, helperKeyPascalString, sizeof(helperKeyPascalString) - 1, kCFStringEncodingMacRoman);
    helperApplication.fCreator = applicationCreatorCode;
    CFStringGetPascalString((CFStringRef)applicationName, helperApplication.name, sizeof(helperApplication.name) - 1, kCFStringEncodingMacRoman);
    helperApplicationSize = sizeof(helperApplication);
    error = ICSetPref(internetConfigInstance, helperKeyPascalString, kICAttrNoChange, &helperApplication, helperApplicationSize);
    switch (error) {
        case noErr:
            break;
        case icPermErr:
            [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"ICSetPref returned permissions when attempting setting key %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error)];
        default:
            [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"ICSetPref returned an error while setting key %@: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), helperKeyString, OANameForInternetConfigErrorCode(error)];
    }
}

- (void)launchURL:(NSString *)urlString;
    // -launchURL: now uses Launch Services rather than Internet Config.  Since this method was really the driving force behind the class (and it's now both simpler to implement and much more robust), we should probably just torch this whole class and replace it with OALaunchServices or something.  But I'm not doing that now, because don't want to change more code than absolutely necessary this close to release.  (I wouldn't have even touched this if Internet Config hadn't been crashing on URLs which encoded Kanji characters.)
{
    OSStatus error;
    CFURLRef url;

    url = CFURLCreateWithString(NULL, (CFStringRef)urlString, NULL);
    error = LSOpenCFURLRef(url, NULL);
    if (url != NULL) // We get NULL for urlStrings like <help:runscript="Essentials:_scripts:PDFFileOpener.as" string="Essentials:SystemOverview:SystemOverview.pdf"> (found on the Help button in <file:///Developer/Documentation/Essentials/SystemOverview/>)
        CFRelease(url);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"Launch Services returned an error while opening URL %@: %@", @"OmniAppKit", [OAInternetConfig bundle], launch services error), urlString, OANameForInternetConfigErrorCode(error)];
}

// Download folder

- (NSString *)downloadFolderPath;
{
    // NSData *data;
    ICAttr attribute;
    struct ICFileSpec *downloadFolderICFileSpecPtr;
    long downloadFolderICFileSpecSize;
    AliasPtr downloadFolderAliasPointer;
    Boolean aliasUpdated;
    FSSpec downloadFolderFSSpec;
    OSStatus error;
    UInt8 path[PATH_MAX + 1];
    FSRef downloadFolderReference;

    // data = [self dataForPreferenceKey:kICDownloadFolder];
    error = ICGetPref(internetConfigInstance, kICDownloadFolder, &attribute, NULL, &downloadFolderICFileSpecSize);
    switch (error) {
        case noErr:
            break;
        case icPrefNotFoundErr:
            // No preference found
            return nil;
        default:
            NSLog(@"OAInternetConfig: ICGetPref returned an error while getting key %s: %@", kICDownloadFolder + 1, OANameForInternetConfigErrorCode(error));
            // No preference found
            return nil;
    }
    downloadFolderICFileSpecPtr = malloc(downloadFolderICFileSpecSize);
    error = ICGetPref(internetConfigInstance, kICDownloadFolder, &attribute, downloadFolderICFileSpecPtr, &downloadFolderICFileSpecSize);
    switch (error) {
        case noErr:
            break;
        case icPrefNotFoundErr:
            // No preference found
            return nil;
        default:
            NSLog(@"OAInternetConfig: ICGetPref returned an error while getting key %s: %@", kICDownloadFolder + 1, OANameForInternetConfigErrorCode(error));
            // No preference found
            return nil;
    }

    downloadFolderAliasPointer = &downloadFolderICFileSpecPtr->alias;
    error = ResolveAlias(NULL, &downloadFolderAliasPointer, &downloadFolderFSSpec, &aliasUpdated);
    if (error != noErr) {
        NSLog(@"-[OAInternetConfig downloadFolderPath]: Error resolving alias: ResolveAlias() failed: %@", OANameForInternetConfigErrorCode(error));
        free(downloadFolderICFileSpecPtr);
        return nil;
    }
    error = FSpMakeFSRef(&downloadFolderFSSpec, &downloadFolderReference);
    if (error != noErr) {
        NSString *name;

        name = (NSString *)CFStringCreateWithPascalString(NULL, downloadFolderFSSpec.name, kCFStringEncodingMacRoman);
        NSLog(@"-[OAInternetConfig downloadFolderPath]: Error converting FSSpec (%d,%d,\"%@\") to FSRef: FSpMakeFSRef() failed: %@", downloadFolderFSSpec.vRefNum, downloadFolderFSSpec.parID, name, OANameForInternetConfigErrorCode(error));
        [name release];
        free(downloadFolderICFileSpecPtr);
        return nil;
    }
    error = FSRefMakePath(&downloadFolderReference, path, sizeof(path));
    if (error != noErr) {
        NSLog(@"-[OAInternetConfig downloadFolderPath]: Error converting FSRef to Path: FSRefMakePath() failed: %@", OANameForInternetConfigErrorCode(error));
        free(downloadFolderICFileSpecPtr);
        return nil;
    }
    return [NSString stringWithCString:path];
}

// Mappings between type/creator codes and filename extensions

- (NSArray *)mapEntries;
{
#warning -mapEntries not finished
    return nil;
}

- (OAInternetConfigMapEntry *)mapEntryForFilename:(NSString *)filename;
{
    OSStatus error;
    ICMapEntry mapEntry;
    Str255 filenamePascalString;

    CFStringGetPascalString((CFStringRef)filename, filenamePascalString, sizeof(filenamePascalString) - 1, kCFStringEncodingMacRoman);
    error = ICMapFilename(internetConfigInstance, filenamePascalString, &mapEntry);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"ICMapFilename returned an error for file %@: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), filename, OANameForInternetConfigErrorCode(error)];
#warning -mapEntryForFilename: not finished
    return nil;
}

- (OAInternetConfigMapEntry *)mapEntryForTypeCode:(long)fileTypeCode creatorCode:(long)fileCreatorCode hintFilename:(NSString *)filename;
{
#warning -mapEntryForTypeCode:creatorCode:hintFilename: not finished
    return nil;
}

- (void)editPreferencesFocusOnKey:(NSString *)keyString;
{
    OSStatus error;
    Str255 keyPascalString;

    if (keyString == nil)
        keyString = @"";
    CFStringGetPascalString((CFStringRef)keyString, keyPascalString, sizeof(keyPascalString) - 1, kCFStringEncodingMacRoman);
    error = ICEditPreferences(internetConfigInstance, keyPascalString);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"Error editing InternetConfig preferences for key %@: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), keyString, OANameForInternetConfigErrorCode(error)];
}

// Low-level access

- (void)beginReadOnlyAccess;
{
    OSStatus error;

    error = ICBegin(internetConfigInstance, icReadOnlyPerm);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"Error opening InternetConfig for read-only access: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), OANameForInternetConfigErrorCode(error)];
}

- (void)beginReadWriteAccess;
{
    OSStatus error;

    error = ICBegin(internetConfigInstance, icReadWritePerm);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"Error opening InternetConfig for read-write access: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), OANameForInternetConfigErrorCode(error)];
}

- (void)endAccess;
{
    OSStatus error;

    error = ICEnd(internetConfigInstance);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"ICEnd returned an error: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), OANameForInternetConfigErrorCode(error)];
}

- (NSArray *)allPreferenceKeys;
{
    OSStatus error;
    NSMutableArray *keys;
    long preferenceIndex, preferenceCount;

    [self beginReadOnlyAccess];
    error = ICCountPref(internetConfigInstance, &preferenceCount);
    if (error != noErr)
        [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"ICCountPref returned an error: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), OANameForInternetConfigErrorCode(error)];
    keys = [NSMutableArray arrayWithCapacity:preferenceCount];
    for (preferenceIndex = 0; preferenceIndex < preferenceCount; preferenceIndex++) {
        Str255 keyPascalString;
        NSString *keyString;

        error = ICGetIndPref(internetConfigInstance, preferenceIndex + 1, keyPascalString);
        if (error != noErr)
            [NSException raise:@"OAInternetConfigException" format:NSLocalizedStringFromTableInBundle(@"ICGetIndPref %d/%d returned an error: %@", @"OmniAppKit", [OAInternetConfig bundle], internet config error), preferenceIndex + 1, preferenceCount, OANameForInternetConfigErrorCode(error)];
        keyString = (NSString *)CFStringCreateWithPascalString(NULL, keyPascalString, kCFStringEncodingMacRoman);
        [keys addObject:keyString];
        [keyString release];
    }
    [self endAccess];
    return keys;
}

- (NSData *)dataForPreferenceKey:(NSString *)preferenceKey;
{
#warning -dataForPreferenceKey: not finished
    return nil;
}

- (void)sendMailViaEntourageTo:(NSString *)receiver carbonCopy:(NSString *)carbonCopy subject:(NSString *)subject body:(NSString *)body;
{
    NSMutableString *script;

    OBPRECONDITION(receiver);

    script = [NSMutableString stringWithFormat:@"tell application \"Microsoft Entourage\"\n make new draft window with properties {"];
    [script appendFormat: @"to recipients:\"%@\"", receiver];

    if (carbonCopy)
        [script appendFormat: @", CC recipients:\"%@\"", carbonCopy];
        
    if (subject)
        [script appendFormat: @", subject:\"%@\"", subject];

    if (body) {
        unsigned int bodyIndex, bodyLength, fragmentLength;

        [script appendString: @", content:\"\" "];
        // AppleScript does not handle string constants longer than 32K.  This can be avoided by using the string concatenation operator.
        // We'll be a little more conservative and only use 16k character strings.
#define APPLE_SCRIPT_MAX_STRING_LENGTH (16*1024)

        bodyLength = [body length];
        bodyIndex = 0;
        while (bodyIndex < bodyLength) {
            NSString *fragment;

            fragmentLength = bodyLength - bodyIndex;
            if (fragmentLength > APPLE_SCRIPT_MAX_STRING_LENGTH)
                fragmentLength = APPLE_SCRIPT_MAX_STRING_LENGTH;

            fragment = [body substringWithRange: (NSRange){bodyIndex, fragmentLength}];
            [script appendFormat: @"& \"%@\" ", fragment];
            bodyIndex += fragmentLength;
        }
    }

    [script appendString: @"}\nactivate\nend tell"];

    // NSLog(@"script = %@", script);

    [OAOSAScript executeScriptString:script];
}

- (void)sendMailVia:(NSString *)mailApp to:(NSString *)receiver carbonCopy:(NSString *)carbonCopy subject:(NSString *)subject body:(NSString *)body;
{
    NSMutableString *script;
    NSMutableArray *arguments;
    NSString *urlCodedString;
    BOOL firstArgument = YES;
    
    OBPRECONDITION(receiver);
        
    script = [NSMutableString stringWithFormat:@"tell application \"%@\"\n %@event GURLGURL%@ \"mailto:", mailApp, [NSString leftPointingDoubleAngleQuotationMarkString], [NSString rightPointingDoubleAngleQuotationMarkString]];
    
    urlCodedString = [NSString encodeURLString:receiver asQuery:NO leaveSlashes:NO leaveColons:NO];
    [script appendString: urlCodedString];

#define NEW_ARG do { \
    [script appendString: firstArgument ? @"?" : @"&"]; \
    firstArgument = NO; \
} while (0)
    
    arguments = [NSMutableArray array];
    if (carbonCopy) {
        NEW_ARG;
        urlCodedString = [NSString encodeURLString:carbonCopy asQuery:NO leaveSlashes:NO leaveColons:NO];
        [script appendFormat: @"cc=%@", urlCodedString];
    }
    
    if (subject) {
        NEW_ARG;
        urlCodedString = [NSString encodeURLString:subject asQuery:NO leaveSlashes:NO leaveColons:NO];
        [script appendFormat: @"subject=%@", urlCodedString];
    }
    
    if (body) {
        unsigned int bodyIndex, bodyLength, fragmentLength;
        NEW_ARG;
        
        [script appendString: @"body=\" "];
        // AppleScript does not handle string constants longer than 32K.  This can be avoided by using the string concatenation operator.  We'll concatenate the body text with the rest of the URL string (to make sure that the rest of the URL doesn't blow the 32K limit in conjunction with the first chunk of body text.
        // We'll be a little more conservative and only use 16k character strings.
#define APPLE_SCRIPT_MAX_STRING_LENGTH (16*1024)

        // URL encode before fragmenting since URL encoding will return longer strings
        urlCodedString = [NSString encodeURLString:body asQuery:NO leaveSlashes:NO leaveColons:NO];
        bodyLength = [urlCodedString length];
        bodyIndex = 0;
        while (bodyIndex < bodyLength) {
            NSString *fragment;
            
            fragmentLength = bodyLength - bodyIndex;
            if (fragmentLength > APPLE_SCRIPT_MAX_STRING_LENGTH)
                fragmentLength = APPLE_SCRIPT_MAX_STRING_LENGTH;
                
            fragment = [urlCodedString substringWithRange: (NSRange){bodyIndex, fragmentLength}];
            [script appendFormat: @"& \"%@\" ", fragment];
            bodyIndex += fragmentLength;
        }
    } else {
        [script appendString:@"\""];
    }
    
        
    [script appendString: @"\nend tell"];
    
    //    NSLog(@"script = %@", script);
    
    [OAOSAScript executeScriptString: script];
}

- (void) launchMailTo:(NSString *) receiver
           carbonCopy:(NSString *) carbonCopy
              subject:(NSString *) subject
                 body:(NSString *) body;
{
    OAInternetConfigInfo appInfo;
    NSString *mailApp;
    BOOL needsWarning;
    
    // Look up the user mail app in InternetConfig (LaunchServices doesn't return app names for mailto urls for some reason).
    appInfo = [self getInfoForScheme: @"mailto"];
    if (!appInfo.valid)
        mailApp = @"Mail";
    else
        mailApp = appInfo.applicationName;
    
    // Need a better, more reliable way of working with a list of known-to-work clients
    needsWarning =  ![[mailApp lowercaseString] hasPrefix:@"mail"] && ([[mailApp lowercaseString] rangeOfString:@"entourage"].length == 0) && ([[mailApp lowercaseString] rangeOfString:@"eudora"].length == 0);
    
/*    
    if (needsWarning && ![[NSUserDefaults standardUserDefaults] boolForKey:@"AlreadyWarnedAboutEmail"]) {
        NSString *warningText, *detailText, *defaultButton, *otherButton, *cancelButton;
        
        warningText = [NSString stringWithFormat:NSLocalizedString(@"CrashCatcher might not work correctly with %@.", email warning message), mailApp];
        detailText = [NSString stringWithFormat:NSLocalizedString(@"The email client you've chosen in System Preferences might not support sending crash reports. We recommend using the built-in Mac OS X Mail application; if you use %@ anyway, the crash report you send might be incomplete (or worse, your email client might crash itself).\n\nIf you choose Cancel, you can still send your crash report by copying and pasting the text into an email message addressed to %@.", email warning details), mailApp, [report emailAddress]];
        defaultButton = NSLocalizedString(@"Send using Mail", email warning default button);
        otherButton = [NSString stringWithFormat:NSLocalizedString(@"Send using %@", email warning message), mailApp];
        cancelButton = NSLocalizedStringFromTableInBundle(@"Cancel", @"Common", [NSApplication bundle], Cancel button on an Alert panel.);
        NSBeginCriticalAlertSheet(warningText, defaultButton, cancelButton, otherButton, [stackTraceTextView window], self, @selector(emailWarningSheetDidEnd:returnCode:contextInfo:), NULL, NULL, detailText);
        return;
    }
*/
    
    if ([[mailApp lowercaseString] rangeOfString:@"entourage"].length != 0) {
        [self sendMailViaEntourageTo:receiver carbonCopy:carbonCopy subject:subject body:body];
    } else {
        [self sendMailVia:mailApp to:receiver carbonCopy:carbonCopy subject:subject body:body];
    }
}

@end

@implementation OAInternetConfig (Private)

- (NSString *)_helperKeyNameForScheme:(NSString *)scheme;
{
    return [helperKeyPrefix stringByAppendingString:scheme];
}

static NSString *OANameForInternetConfigErrorCode(OSStatus errorCode)
{
    switch (errorCode) {
        case icPrefNotFoundErr: return @"icPrefNotFoundErr";
        case icPermErr: return @"icPermErr";
        case icPrefDataErr: return @"icPrefDataErr";
        case icInternalErr: return @"icInternalErr";
        case icTruncatedErr: return @"icTruncatedErr";
        case icNoMoreWritersErr: return @"icNoMoreWritersErr";
        case icNothingToOverrideErr: return @"icNothingToOverrideErr";
        case icNoURLErr: return @"icNoURLErr";
        case icConfigNotFoundErr: return @"icConfigNotFoundErr";
        case icConfigInappropriateErr: return @"icConfigInappropriateErr";
        case icProfileNotFoundErr: return @"icProfileNotFoundErr";
        case icTooManyProfilesErr: return @"icTooManyProfilesErr";
        default: return [NSString stringWithFormat:@"<error code %d>", errorCode];
    }
}

@end
