// Copyright 1997-2002 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniAppKit/OAApplication.h>

#import <Foundation/Foundation.h>
#import <AppKit/AppKit.h>
#import <OmniBase/OmniBase.h>
#import <OmniFoundation/OmniFoundation.h>

#import <Carbon/Carbon.h>

#import "NSView-OAExtensions.h"
#import "NSImage-OAExtensions.h"
#import "OAAppkitQueueProcessor.h"
#import "OAInspector.h"

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/Frameworks/OmniAppKit/OAApplication.m,v 1.62 2002/03/09 01:53:53 kc Exp $")

@interface NSObject (OAAppDelegateExtensions)
- (void)openAddressWithString:(NSString *)addressString;
    // This method is known to exist in OmniWeb, and is used to see if we're running in OmniWeb.
    // If we're running in OmniWeb, we use this method to show help content without leaving the app.
@end

@interface OAApplication (Private)
- (void)processMouseButtonsChangedEvent:(NSEvent *)event;
+ (void)_setupDynamicMenus:(NSMenu *)aMenu;
+ (void)_makeDynamicItemGroupInMenu:(NSMenu *)aMenu forRangeOfItems:(NSRange)aRange;
- (void)_runNewVersionDialog:(NSDictionary *)versionInfo;
@end

@implementation OAApplication

+ (void)setupOmniApplication;
{
    [OBObject self]; // Trigger +[OBPostLoader processClasses]
    
    // make this image available to client nibs and whatnot
    [NSImage imageNamed:@"OAHelpIcon" inBundleForClass:[OAApplication class]];
}

+ (NSApplication *)sharedApplication;
{
    static OAApplication *omniApplication = nil;

    if (omniApplication)
        return omniApplication;

    omniApplication = (id)[super sharedApplication];
    [OAApplication setupOmniApplication];
    return omniApplication;
}

- (void)finishLaunching;
{
    [[OFController sharedController] addObserver:[OAApplication class]];
    OBASSERT(softwareUpdate == nil);
    softwareUpdate = [[OFSoftwareUpdateChecker alloc] initWithTarget:self action:@selector(_runNewVersionDialog:)];
    [super finishLaunching];
}

- (void)run;
{
    BOOL stillRunning = YES;

    exceptionCount = 0;
    exceptionCheckpointDate = [[NSDate alloc] init];
    do {
        NS_DURING {
            [super run];
            stillRunning = NO;
        } NS_HANDLER {
            if (++exceptionCount % 300 == 0) {
                if ([exceptionCheckpointDate timeIntervalSinceNow] > -3.0)
                    stillRunning = NO; // 300 exceptions in 3 seconds: abort
                [exceptionCheckpointDate release];
                exceptionCheckpointDate = [[NSDate alloc] init];
                exceptionCount = 0;
            }
            if (localException) {
                if (_appFlags._hasBeenRun)
                    [self handleRunException:localException];
                else
                    [self handleInitException:localException];
            }
        } NS_ENDHANDLER;
    } while (stillRunning && _appFlags._hasBeenRun);
}

#define MAXIMUM_LINE_FACTOR 12.0
#define PAGE_FACTOR MAXIMUM_LINE_FACTOR * 2.0 * 2.0 * 2.0
#define ACCELERATION 2.0
#define MAX_SCALE_SETTINGS 12

static struct {
    float targetScrollFactor;
    float timeSinceLastScroll;
} mouseScaling[MAX_SCALE_SETTINGS] = {
    {2.0, 0.0}
};

static void OATargetScrollFactorReadFromDefaults(void)
{
    NSArray *values;
    unsigned int settingIndex, valueCount;

    values = [[NSUserDefaults standardUserDefaults] arrayForKey:@"OAScrollWheelTargetScrollFactor"];
    if (values == nil)
        return;
    valueCount = [values count];
    for (settingIndex = 0; settingIndex < MAX_SCALE_SETTINGS; settingIndex++) {
        unsigned int factorValueIndex;
        float factor, cutoff;

        factorValueIndex = settingIndex * 2;
        factor = factorValueIndex < valueCount ? [[values objectAtIndex:factorValueIndex] floatValue] : 0.0;
        cutoff = factorValueIndex + 1 < valueCount ? (1.0 / [[values objectAtIndex:factorValueIndex + 1] floatValue]) : 0.0;
        mouseScaling[settingIndex].targetScrollFactor = factor;
        mouseScaling[settingIndex].timeSinceLastScroll = cutoff;
    }
}

static float OATargetScrollFactorForTimeInterval(NSTimeInterval timeSinceLastScroll)
{
    static BOOL alreadyInitialized = NO;
    unsigned int mouseScalingIndex;

    if (!alreadyInitialized) {
        OATargetScrollFactorReadFromDefaults();
        alreadyInitialized = YES;
    }
    for (mouseScalingIndex = 0;
         mouseScalingIndex < MAX_SCALE_SETTINGS && MAX(0.0, timeSinceLastScroll) < mouseScaling[mouseScalingIndex].timeSinceLastScroll;
         mouseScalingIndex++) {
    }

    return mouseScaling[mouseScalingIndex].targetScrollFactor;
}

static float OAScrollFactorForWheelEvent(NSEvent *event)
{
    static NSTimeInterval lastScrollWheelTimeInterval = 0.0;
    static float scrollFactor = 100.0;
    NSTimeInterval timestamp;
    NSTimeInterval timeSinceLastScroll;
    float deltaY;
    float targetScrollFactor;
    
    timestamp = [event timestamp];
    timeSinceLastScroll = timestamp - lastScrollWheelTimeInterval;
    targetScrollFactor = OATargetScrollFactorForTimeInterval(timeSinceLastScroll);
    lastScrollWheelTimeInterval = timestamp;
    deltaY = -[event deltaY];
    if (scrollFactor == targetScrollFactor) {
        // Do nothing
    } else if (timeSinceLastScroll > 0.5) {
        // If it's been more than half a second, just start over at the target factor
        scrollFactor = targetScrollFactor;
    } else if (scrollFactor * (1.0 / ACCELERATION) > targetScrollFactor) {
        // Reduce our scroll factor
        scrollFactor *= (1.0 / ACCELERATION);
    } else if (scrollFactor * ACCELERATION < targetScrollFactor) {
        // Increase our scroll factor
        scrollFactor *= ACCELERATION;
    } else {
        // The target is near, just jump to it
        scrollFactor = targetScrollFactor;
    }
    return scrollFactor;
}

#define OASystemDefinedEvent_MouseButtonsChangedSubType 7

- (void)sendEvent:(NSEvent *)event;
{
    // The -timestamp method on NSEvent doesn't seem to return an NSTimeInterval based off the same reference date as NSDate (which is what we want).
    lastEventTimeInterval = [NSDate timeIntervalSinceReferenceDate];

    NS_DURING {
        switch ([event type]) {
            case NSSystemDefined:
                if ([event subtype] == OASystemDefinedEvent_MouseButtonsChangedSubType)
                    [self processMouseButtonsChangedEvent:event];
                [super sendEvent:event];
                break;
            case NSScrollWheel:
            {
                NSView *contentView;
                NSView *viewUnderMouse;
                NSScrollView *scrollView;
                BOOL scrollWheelButtonIsDown;
    
                contentView = [[event window] contentView];
                viewUnderMouse = [contentView hitTest:[contentView convertPoint:[event locationInWindow] fromView:nil]];
                scrollView = [viewUnderMouse enclosingScrollView];
                scrollWheelButtonIsDown = [self scrollWheelButtonIsDown];
                if (scrollView == nil) {
                    // Punt
                    [super sendEvent:event];
                    if (scrollWheelButtonIsDown) {
                        // Multiply the strength of the original event
                        [super sendEvent:event];
                        [super sendEvent:event];
                        [super sendEvent:event];
                        [super sendEvent:event];
                    }
                } else {
                    float deltaY;

                    deltaY = -[event deltaY];
                    if (scrollWheelButtonIsDown) {
                        [scrollView scrollDownByPages:deltaY];
                    } else {
                        float scrollFactor;

                        scrollFactor = OAScrollFactorForWheelEvent(event);
                        if (scrollFactor >= PAGE_FACTOR)
                            [scrollView scrollDownByPages:deltaY];
                        else
                            [scrollView scrollDownByLines:MIN(scrollFactor, MAXIMUM_LINE_FACTOR) * deltaY];
                    }
                }
                break;
            }
            default:
                [super sendEvent:event];
                break;
        }
    } NS_HANDLER {
        if ([[localException name] isEqualToString:NSAbortModalException] || [[localException name] isEqualToString:NSAbortPrintingException])
            [localException raise];
        [self handleRunException:localException];
    } NS_ENDHANDLER;
}

- (void)handleInitException:(NSException *)anException;
{
    id delegate;

    delegate = [self delegate];
    if ([delegate respondsToSelector:@selector(handleInitException:)]) {
        [delegate handleInitException:anException];
    } else {
        NSLog(@"%@", [anException reason]);
    }
}

- (void)handleRunException:(NSException *)anException;
{
    id delegate;

    delegate = [self delegate];
    if ([delegate respondsToSelector:@selector(handleRunException:)]) {
        [delegate handleRunException:anException];
    } else {
        NSLog(@"%@", [anException reason]);
        NSRunAlertPanel(nil, @"%@", nil, nil, nil, [anException reason]);
    }
}

- (NSTimeInterval)lastEventTimeInterval;
{
    return lastEventTimeInterval;
}

- (BOOL)mouseButtonIsDownAtIndex:(unsigned int)mouseButtonIndex;
{
    return (mouseButtonState & (1 << mouseButtonIndex)) != 0;
}

- (BOOL)scrollWheelButtonIsDown;
{
    return [self mouseButtonIsDownAtIndex:2];
}

- (void)showHelpURL:(NSString *)helpURL;
{
    id applicationDelegate;
    
    // NB: See similar code in _runNewVersionDialog:
    
    applicationDelegate = [NSApp delegate];
    if ([applicationDelegate respondsToSelector:@selector(openAddressWithString:)]) {
        // We're presumably in OmniWeb, in which case we display our help internally
        NSString *omniwebHelpBaseURL = @"omniweb:/Help/";
        [applicationDelegate openAddressWithString:[omniwebHelpBaseURL stringByAppendingString:helpURL]];
    } else if ([[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleHelpBookName"]) {
        // We've got Apple Help
        NSString *bookName = [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleHelpBookName"];
        OSStatus err = AHGotoPage((CFStringRef)bookName, (CFStringRef)helpURL, NULL);
        if (err != noErr)	
            NSLog(@"Apple Help error: %d", err);
    } else {
        // We can let the system decide who to open the URL with
        [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:helpURL]];
    }
}

// Actions

- (IBAction)closeAllMainWindows:(id)sender;
{
    NSArray *windows;
    unsigned int windowIndex, windowCount;
    
    windows = [[NSArray alloc] initWithArray:[self orderedWindows]];
    windowCount = [windows count];
    for (windowIndex = 0; windowIndex < windowCount; windowIndex++) {
        NSWindow *window;

        window = [windows objectAtIndex:windowIndex];
        if ([window canBecomeMainWindow])
            [window performClose:nil];
    }
    [windows release];
}

- (IBAction)miniaturizeAll:(id)sender;
{
    NSArray *windows;
    unsigned int windowIndex, windowCount;
    
    windows = [[NSArray alloc] initWithArray:[self orderedWindows]];
    windowCount = [windows count];
    for (windowIndex = 0; windowIndex < windowCount; windowIndex++) {
        NSWindow *window;

        window = [windows objectAtIndex:windowIndex];
        [window performMiniaturize:nil];
    }
    [windows release];
}

- (IBAction)cycleToNextMainWindow:(id)sender;
{
    NSWindow *mainWindow;
    NSArray *orderedWindows;
    unsigned int windowIndex, windowCount;
    
    mainWindow = [NSApp mainWindow];
    orderedWindows = [NSApp orderedWindows];
    windowCount = [orderedWindows count];
    for (windowIndex = 0; windowIndex < windowCount; windowIndex++) {
        NSWindow *window;

        window = [orderedWindows objectAtIndex:windowIndex];
        if (window != mainWindow && [window canBecomeMainWindow] && ![NSStringFromClass([window class]) isEqualToString:@"NSDrawerWindow"]) {
            [window makeKeyAndOrderFront:nil];
            [mainWindow orderBack:nil];
            return;
        }
    }
    // There's one (or less) window which can potentially be main, make it key and bring it forward.
    [mainWindow makeKeyAndOrderFront:nil];
}

- (IBAction)cycleToPreviousMainWindow:(id)sender;
{
    NSWindow *mainWindow;
    NSArray *orderedWindows;
    unsigned int windowIndex;
    
    mainWindow = [NSApp mainWindow];
    orderedWindows = [NSApp orderedWindows];
    windowIndex = [orderedWindows count];
    while (windowIndex--) {
        NSWindow *window;

        window = [orderedWindows objectAtIndex:windowIndex];
        if (window != mainWindow && [window canBecomeMainWindow] && ![NSStringFromClass([window class]) isEqualToString:@"NSDrawerWindow"]) {
            [window makeKeyAndOrderFront:nil];
            return;
        }
    }
    // There's one (or less) window which can potentially be main, make it key and bring it forward.
    [mainWindow makeKeyAndOrderFront:nil];
}

- (IBAction)showInspectorPanel:(id)sender;
{
    [OAInspector showInspector];
}

// Check for new version of this application on Omni's web site. Triggered by direct user action.
- (IBAction)checkForNewVersion:(id)sender;
{
    [self checkForNewVersionFromWindow:nil];
}

- (void)checkForNewVersionFromWindow:(NSWindow *)aWindow;
{
    NSString *primaryMessage = nil;
    NSString *secondaryMessage = nil;
    
    NS_DURING {
        BOOL didSomething = [softwareUpdate checkSynchronously];
        
        if (!didSomething) {
            primaryMessage = NSLocalizedStringFromTableInBundle(@"There is no new version of %@ available.", @"OmniAppKit", [OAApplication bundle], text of dialog box informing user that no new software is available);
            primaryMessage = [NSString stringWithFormat:primaryMessage, [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleName"]];
            secondaryMessage = @"";
        }
    } NS_HANDLER {
        NSLog(@"Software update exception: %@", localException);
        primaryMessage = NSLocalizedStringFromTableInBundle(@"Unable to check for new software.", @"OmniAppKit", [OAApplication bundle], text of dialog box informing user that software update failed -  probably some sort of network error);
        secondaryMessage = [NSString stringWithFormat:NSLocalizedStringFromTableInBundle(@"Attempting to check for updates failed due to the following error:\n\n%@", @"OmniAppKit", [OAApplication bundle], informational text of dialog box informing user that software update failed), [localException reason]];
    } NS_ENDHANDLER;
    
    if (primaryMessage != nil) {
        NSString *okButton;
        
        okButton = NSLocalizedStringFromTableInBundle(@"OK", @"OmniAppKit", [OAApplication bundle], button on dialog box informing user that software update did not discover new software);
        NSRunInformationalAlertPanel(primaryMessage, secondaryMessage, okButton, NULL, NULL);
    }
}

// OFController observer informal protocol

+ (void)controllerStartedRunning:(OFController *)controller;
{
    [self _setupDynamicMenus:[NSApp mainMenu]];
}

@end

@implementation OAApplication (Private)

- (void)processMouseButtonsChangedEvent:(NSEvent *)event;
{
    mouseButtonState = [event data2];
}

+ (void)_setupDynamicMenus:(NSMenu *)aMenu;
{
    unsigned int itemIndex, itemCount;
    NSRange dynamicItemGroupRange;

    dynamicItemGroupRange = NSMakeRange(0, 0);
    itemCount = [aMenu numberOfItems];
    for (itemIndex = 0; itemIndex < itemCount; itemIndex++) {
        NSMenuItem *currentItem;

        currentItem = [aMenu itemAtIndex:itemIndex];
        if ([currentItem hasSubmenu])
            [self _setupDynamicMenus:[currentItem submenu]];
        if ([currentItem tag] == 1) {
            if (dynamicItemGroupRange.length == 0) // If we don't have a group open
                dynamicItemGroupRange.location = itemIndex; // Start the group here
            dynamicItemGroupRange.length++; // Extend the group
        } else if (dynamicItemGroupRange.length != 0 && [currentItem tag] == 0) {
            // We've reached the end of the group, process it
            [self _makeDynamicItemGroupInMenu:aMenu forRangeOfItems:dynamicItemGroupRange];
            dynamicItemGroupRange.length = 0;
        }
    }

    // If we ended with a dynamic group, process it
    if (dynamicItemGroupRange.length != 0)
        [self _makeDynamicItemGroupInMenu:aMenu forRangeOfItems:dynamicItemGroupRange];
}

+ (void)_makeDynamicItemGroupInMenu:(NSMenu *)aMenu forRangeOfItems:(NSRange)aRange;
{
    extern MenuRef _NSGetCarbonMenu(NSMenu *);
    MenuRef menu;
    int item;
    
    menu = _NSGetCarbonMenu(aMenu);
    aRange.location++; // Carbon menu indices include the menu itself at index 0:  skip past it
    for (item = aRange.location; item < aRange.location + aRange.length; item++)
        ChangeMenuItemAttributes(menu, item, kMenuItemAttrDynamic, 0);
}

- (void)_runNewVersionDialog:(NSDictionary *)versionInfo
{
    NSString *title, *message, *yesButton, *noButton, *dlButton;
    NSString *thisApp, *displayVersion;
    NSString *downloadPage, *directDownload;
    NSString *urlToRetrieve = nil;
    int userChoice;
    
    // TODO --- This depends on knowing the dictionary keys retrieved from the server. We should decouple this method interface from the plist format.
    
    thisApp = [versionInfo objectForKey:@"displayName"];
    if (!thisApp)
        thisApp = [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleName"];
    displayVersion = [versionInfo objectForKey:@"displayVersion"];
    downloadPage = [versionInfo objectForKey:@"downloadPage"];
    directDownload = [versionInfo objectForKey:@"directDownload"];
    
    title = [NSString stringWithFormat:NSLocalizedStringFromTableInBundle(@"A new version of %@ is now available.", @"OmniAppKit", [OAApplication bundle], primary message of dialog box asking user to download new version), thisApp];
    message = @"";
    if (displayVersion) {
        NSString *text = NSLocalizedStringFromTableInBundle(@"The latest version of %@ is %@. ", @"OmniAppKit", [OAApplication bundle], informational text of dialog box asking user to download new version - displays version string of new version - should end with whitespace since another sentence may be appended.);
        message = [message stringByAppendingFormat:text, thisApp, displayVersion];
    }
    if (downloadPage) {
        NSString *text = NSLocalizedStringFromTableInBundle(@"Choose \"More Info\" to open a web page from which you can download or get information about this new version. ", @"OmniAppKit", [OAApplication bundle], informational text of dialog box asking user to download new version - should end with whitespace since another sentence may be appended.);
        message = [message stringByAppendingString:text];
    }
    if ([[NSUserDefaults standardUserDefaults] boolForKey:@"AutomaticSoftwareUpdateCheckEnabled"]) {
        NSString *text = NSLocalizedStringFromTableInBundle(@"To disable automatic update checking, see the \"Update\" pane in %@'s Preferences window.", @"OmniAppKit", [OAApplication bundle], informational text of dialog box asking user to download new version - appended when the dialog is displayed because of a scheduled software update check which happened in the background);
        message = [message stringByAppendingFormat:text, thisApp];
    }
    yesButton = NSLocalizedStringFromTableInBundle(@"More Info...", @"OmniAppKit", [OAApplication bundle], button on software update dialog box - this button opens a web page with information about the new version);
    noButton = NSLocalizedStringFromTableInBundle(@"Update Later", @"OmniAppKit", [OAApplication bundle], button on software update dialog box - this button dismisses the dialog without doing anything);
    dlButton = NSLocalizedStringFromTableInBundle(@"Download now", @"OmniAppKit", [OAApplication bundle], button on software update dialog box - this button downloads the new version immediately);
    
    
    /* Run an alert panel */
    
    if (downloadPage && directDownload) {
        userChoice = NSRunInformationalAlertPanel(title, message, yesButton, noButton, dlButton);
        if (userChoice == NSAlertDefaultReturn)
            urlToRetrieve = downloadPage;
        else if (userChoice == NSAlertOtherReturn)
            urlToRetrieve = directDownload;
    } else if (downloadPage && !directDownload) {
        userChoice = NSRunInformationalAlertPanel(title, message, yesButton, noButton, NULL);
        if (userChoice == NSAlertDefaultReturn)
            urlToRetrieve = downloadPage;
    } else if (!downloadPage && directDownload) {
        userChoice = NSRunInformationalAlertPanel(title, message, dlButton, noButton, NULL);
        if (userChoice == NSAlertDefaultReturn)
            urlToRetrieve = directDownload;
    } else {
        /* Okay, apparently we know about the new version but don't know how to get it. */
        NSString *sheeshButton = NSLocalizedStringFromTableInBundle(@"OK", @"OmniAppKit", [OAApplication bundle], button on software update dialog box - for when we know there's a new version but can't offer to download it);
        NSRunInformationalAlertPanel(title, message, sheeshButton, NULL, NULL);
    }
    
    
    if (urlToRetrieve != nil) {
        id applicationDelegate;
    
        applicationDelegate = [NSApp delegate];
        // Code stolen from -showHelpURL:
        if ([applicationDelegate respondsToSelector:@selector(openAddressWithString:)]) {
            // We're presumably in OmniWeb, in which case we display web pages internally
            [applicationDelegate openAddressWithString:urlToRetrieve];
        } else {
            // We can let the system decide who to open the URL with
            [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:urlToRetrieve]];
        }
    }
}

@end
