// Copyright 2002 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OAToolbarWindowController.h"

#import <AppKit/AppKit.h>
#import <OmniAppKit/OmniAppKit.h>
#import <OmniBase/OmniBase.h>

#import "OAScriptToolbarHelper.h"

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/Frameworks/OmniAppKit/OAToolbarWindowController.m,v 1.6 2002/03/09 01:53:53 kc Exp $")

@implementation OAToolbarWindowController

static NSMutableDictionary *toolbarItemInfo = nil;
static NSMutableDictionary *allowedToolbarItems = nil;
static NSMutableDictionary *defaultToolbarItems = nil;
static NSMutableDictionary *helpersByExtension = nil;

+ (void)initialize;
{
    OBINITIALIZE;
    
    toolbarItemInfo = [[NSMutableDictionary alloc] init];
    allowedToolbarItems = [[NSMutableDictionary alloc] init];
    defaultToolbarItems = [[NSMutableDictionary alloc] init];
    helpersByExtension = [[NSMutableDictionary alloc] init];
    
    [self registerToolbarHelper:[[OAScriptToolbarHelper alloc] init]];
}

+ (void)loadToolbarNamed:(NSString *)toolbarName;
{
    NSDictionary *toolbarPropertyList;

    if ([allowedToolbarItems objectForKey:toolbarName] != nil)
        return;

    
    toolbarPropertyList = [NSDictionary dictionaryWithContentsOfFile:[[NSBundle bundleForClass:self] pathForResource:toolbarName ofType:@"toolbar"]];
    [toolbarItemInfo setObject:[toolbarPropertyList objectForKey:@"itemInfoByIdentifier"] forKey:toolbarName];
    [allowedToolbarItems setObject:[toolbarPropertyList objectForKey:@"allowedItemIdentifiers"] forKey:toolbarName];
    [defaultToolbarItems setObject:[toolbarPropertyList objectForKey:@"defaultItemIdentifiers"] forKey:toolbarName];
}

+ (void)registerToolbarHelper:(NSObject <OAToolbarHelper> *)helperObject;
{
    [helpersByExtension setObject:helperObject forKey:[helperObject itemIdentifierExtension]];
}

- (NSDictionary *)itemInfoForIdentifier:(NSString *)itemIdentifier;
{
    NSString *name;
    
    name = [self toolbarConfigurationName];
    [isa loadToolbarNamed:name];
    return [[toolbarItemInfo objectForKey:name] objectForKey:itemIdentifier];
}

//
// Implement in subclasses
//

- (NSString *)toolbarConfigurationName;
{
    return @"Toolbar";
}

- (NSString *)toolbarIdentifier;
{
    return [self toolbarConfigurationName];
}

//
// Setup the toolbar and handle its delegate messages
//

- (void)windowDidLoad;
{
    [isa loadToolbarNamed:[self toolbarConfigurationName]];
    toolbar = [[NSToolbar alloc] initWithIdentifier:[self toolbarIdentifier]];
    [toolbar setAllowsUserCustomization:YES];
    [toolbar setAutosavesConfiguration:YES];
    [toolbar setDelegate:self];
    [[self window] setToolbar:toolbar];
}

- (NSToolbarItem *)toolbar:(NSToolbar *)toolbar itemForItemIdentifier:(NSString *)itemIdentifier willBeInsertedIntoToolbar:(BOOL)willInsert;
{
    OAToolbarItem *newItem;
    NSDictionary *itemInfo;
    NSString *itemImageName;
    NSArray *sizes;
    NSObject <OAToolbarHelper> *helper = nil;
    NSString *extension, *value;
    NSImage *itemImage = nil;
    
    // Always use OAToolbarItem since we can't know up front whether we'll need a delegate or not.
    newItem = [[[OAToolbarItem alloc] initWithItemIdentifier:itemIdentifier] autorelease];
    
    if ((extension = [itemIdentifier pathExtension])) 
        helper = [helpersByExtension objectForKey:extension];

    if (helper) {
        NSString *nameWithoutExtension = [itemIdentifier stringByDeletingPathExtension];

        itemInfo = [[toolbarItemInfo objectForKey:[self toolbarConfigurationName]] objectForKey:[helper templateItemIdentifier]];

        if ((value = [itemInfo objectForKey:@"label"]))
            [newItem setLabel:[NSString stringWithFormat:value, nameWithoutExtension]];
        if ((value = [itemInfo objectForKey:@"paletteLabel"]))
            [newItem setPaletteLabel:[NSString stringWithFormat:value, nameWithoutExtension]];
        if ((value = [itemInfo objectForKey:@"toolTip"]))
            [newItem setToolTip:[NSString stringWithFormat:value, nameWithoutExtension]];
        
        // let custom item have custom image    
        if (OACurrentControlTint() == OAGraphiteTint)
            itemImage = [NSImage imageNamed:[itemIdentifier stringByAppendingString:@"Graphite"]];
        if (!itemImage)
            itemImage = [NSImage imageNamed:itemIdentifier];
    } else {
        itemInfo = [[toolbarItemInfo objectForKey:[self toolbarConfigurationName]] objectForKey:itemIdentifier];

        if ((value = [itemInfo objectForKey:@"label"]))
            [newItem setLabel:value];
        if ((value = [itemInfo objectForKey:@"paletteLabel"]))
            [newItem setPaletteLabel:value];
        if ((value = [itemInfo objectForKey:@"toolTip"]))
            [newItem setToolTip:value];
    }
    
    if ((value = [itemInfo objectForKey:@"customView"])) {
        // customView should map to a method or ivar on a subclass
        NSView *customView;
        NSRect frame;
        
        customView = [self valueForKey: value];
        OBASSERT(customView);
        [newItem setView: customView];
        
        // We have to provide validation for items with custom views.
        [newItem setDelegate: self];
        
        // Default to having the min size be the current size of the view and the max size unbounded.
        frame = [customView frame];
        [newItem setMinSize: frame.size];
    }
    
    [newItem setTarget:self];
    if ((value = [itemInfo objectForKey:@"target"])) {
        if ([value isEqualToString:@"firstResponder"])
            [newItem setTarget:nil];
    }
    if ((value = [itemInfo objectForKey:@"action"]))
        [newItem setAction:NSSelectorFromString(value)];
        
    sizes = [itemInfo objectForKey:@"minSize"];
    if (sizes)
        [newItem setMinSize:NSMakeSize([[sizes objectAtIndex:0] floatValue], [[sizes objectAtIndex:1] floatValue])];
    sizes = [itemInfo objectForKey:@"maxSize"];
    if (sizes)
        [newItem setMaxSize:NSMakeSize([[sizes objectAtIndex:0] floatValue], [[sizes objectAtIndex:1] floatValue])];
    
    if (!itemImage && ((itemImageName = [itemInfo objectForKey:@"imageName"]))) {
        if (OACurrentControlTint() == OAGraphiteTint)
            itemImage = [NSImage imageNamed:[itemImageName stringByAppendingString:@"Graphite"]];
        if (!itemImage)
            itemImage = [NSImage imageNamed:itemImageName];
    }
    
    if (itemImage)
        [newItem setImage:itemImage];
        
    if (helper)
        [helper finishSetupForItem:newItem];
        
    return newItem;
}

- (NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar *)toolbar;
{
    NSEnumerator *enumerator;
    NSObject <OAToolbarHelper> *helper;
    NSMutableArray *results;
    
    results = [NSMutableArray arrayWithArray:[allowedToolbarItems objectForKey:[self toolbarConfigurationName]]];
    enumerator = [helpersByExtension objectEnumerator];
    while ((helper = [enumerator nextObject])) {
        int itemIndex;
        
        itemIndex = [results indexOfObject:[helper templateItemIdentifier]];
        if (itemIndex == NSNotFound)
            continue;
        [results replaceObjectsInRange:NSMakeRange(itemIndex, 1) withObjectsFromArray:[helper allowedItems]];
    }
    return results;
}

- (NSArray *)toolbarDefaultItemIdentifiers:(NSToolbar *)toolbar;
{
    return [defaultToolbarItems objectForKey:[self toolbarConfigurationName]];
}

- (void)toolbarWillAddItem:(NSNotification *)notification;
{
}

- (void)toolbarDidRemoveItem:(NSNotification *)notification;
{
}

- (BOOL)validateToolbarItem:(NSToolbarItem *)theItem;
{
    return YES;
}

@end
